! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: eimsimreliability_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This is a collection of routines used to reduce simulations data.
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module eimsimreliability_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.

  use dal
  use param  
  use errorHandling

  use array_utils	! for minloc1d
  use dal_aux	! for addOrOpenColumn()
!  use histogram_utils	! for makeHistogram().
use l_histogram_utils
!  use pgplot_aux
use l_pgplot_aux
  use ssc_misc	! for fracVal().

  implicit none 

  private 

  ! Command-line parameters:
  !
  character(256) ::&
    histoSetName='',&
    detSrcSetName='',&
    plot1FileName='',&
    plot2FileName='',&
    plot3FileName='',&
    plot4FileName=''
  real(single) ::&
    nullProbCutoff=0.0,&
    histoBinSize=0.0,&
    minHistoX=0.0
  logical ::&
    makePlots=.false.

character(20) :: pgDev = '/cps'

  public eimsimreliability

contains 

  !---------------------------------------------------------------------
  subroutine eimsimreliability
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    call readParameters

    call calcReliabilityHistogramPart1
    call calcReliabilityHistogramPart2

    if (makePlots) then
      call message(AppMsg, Verbose, "Constructing the plots.")

      ! Make a scatter plot of x=DPNUL_MID, y=N_BAD, y_err=N_BAD_ERR; both logarithmic axes. The number of false detections N_BAD should be proportional to the null probability DPNUL, thus this log-log plot should have slope of 1. The proportionality is the number of 'beams' in the image.
      !
!print *, 'plot 1'
      call makePlot1

      ! Now do the same, but a line plot from the cumulative integral of N_BAD. Here x=DPNUL_INT and y=N_BAD_INT; both logarithmic axes.
      !
!print *, 'plot 2'
      call makePlot2

      ! Plot 3 is a simple comparison between the fluxes of detected and simulated sources. Three plots are superimposed as follows:
      !	1:	x=SIM_FLUX	y=FLUX	y_err=FLUX_ERR		red point+err
      !	2:	A straight line of slope 1			blue line
      !	3:	x=SIM_FLUX	y=FLUX	y_err=none		black point
      !
!print *, 'plot 3'
      call makePlot3

      ! Plot 4 shows how the 'match' probability varies with flux.
      !		x=SIM_FLUX	y=MATCH_PNULL	point plot, both axes logarithmic.
      !
!print *, 'plot 4'
      call makePlot4
    end if

  end subroutine eimsimreliability

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

    histoBinSize = realParameter('histobinsize')
    histoSetName = stringParameter('histoset')
    detSrcSetName = stringParameter('detsrclistset')
    nullProbCutoff = realParameter('probcutoff')
    minHistoX = realParameter('minhistoprob')

    makePlots = booleanParameter('withplots')
    if (makePlots) then
      plot1FileName = stringParameter('plot1file')
      plot2FileName = stringParameter('plot2file')
      plot3FileName = stringParameter('plot3file')
      plot4FileName = stringParameter('plot4file')
    end if

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine calcReliabilityHistogramPart1
  !---------------------------------------------------------------------
  ! Makes a single fits file that contains some histograms (usually over
  ! DET_ML). The column names are as follows:
  !
  !	N_ALL		Number sources detected.
  !	N_ALL_ERR	Square root of N_ALL.
  !	N_ALL_INT	Integrated value of N_ALL.
  !	N_TRUE		Number of reliable detections.
  !	N_TRUE_ERR	Square root of N_TRUE.
  !	N_TRUE_INT	Integrated value of N_TRUE.
  !	N_BAD		Number of unreliable detections.
  !	N_BAD_ERR	Square root of N_BAD.
  !	N_BAD_INT	Integrated value of N_BAD.
  !	DPNUL_INT	Lower edge of the bin.
  !	DPNUL_MID	Middle of the bin.
  !	DPNUL_HI	Upper edge of the bin.
  !
  ! NOTE! If the source list has no COMPARED keyword, or if this == .false., all sources are considered to be 'bad', and the *_ALL* and *_TRUE* columns are omitted.
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) :: detSrcSet, histoSet
    type(TableT) :: detSrcTab, histoTab
    integer ::&
      numDetSources, numHistoBins, bin, numUseableSources, numReliableSources,&
      numUnreliableSources, dummyUnderflows, dummyOverflows
    type(ColumnT) ::&
      detecNullProbCol, simInverseSensyCol, matchNullProbCol,&
      nAllCol, nAllErrCol, nAllIntCol, nBadCol, nBadErrCol, nBadIntCol,&
      nTruCol, nTruErrCol, nTruIntCol, detectNullProbLoCol,&
      detectNullProbMidCol, detectNullProbHiCol
    real(single), pointer ::&
      detecNullProb(:), simInverseSensy(:), matchNullProb(:), nAllErr(:),&
      binEdges(:), nBad(:), nBadInt(:), nBadErr(:), nTru(:), nTruInt(:), nTruErr(:),&
      detectNullProbLo(:), detectNullProbMid(:), detectNullProbHi(:)
    real(single), allocatable ::&
      log10DetecNullProb(:), dummyWeights(:)
    logical, allocatable :: selectionMask(:)
    type(weightedHistoSingleType), allocatable :: detTotalHisto(:)
    integer(int32), pointer :: nAll(:), nAllInt(:)
    integer(int32), allocatable :: detBadHistoInt(:)
    logical :: withSimSources

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose&
      , "Starting the routine which makes the basic histogram of det source flux.")

    detSrcSet = dataSet(detSrcSetName, READ)
    detSrcTab = table(detSrcSet, 'SRCLIST')

    withSimSources = .false. ! default
    if (hasAttribute(detSrcTab, 'COMPARED')) then
      if (booleanAttribute(detSrcTab, 'COMPARED')) withSimSources = .true.
    end if

    numDetSources = numberOfRows(detSrcTab)
    allocate(selectionMask(numDetSources))

    detecNullProbCol   = column(detSrcTab, 'DETEC_PNULL')
    detecNullProb   => real32Data(detecNullProbCol)

    if (withSimSources) then
      simInverseSensyCol = column(detSrcTab, 'SIM_INV_SENSY')
      matchNullProbCol   = column(detSrcTab, 'MATCH_PNULL')

      simInverseSensy => real32Data(simInverseSensyCol)
      matchNullProb   => real32Data(matchNullProbCol)

!      selectionMask = (simInverseSensy>0.0 .and. detecNullProb>0.0)
      selectionMask = (simInverseSensy>0.0 .and. detecNullProb>minHistoX)
    else
!      selectionMask = (detecNullProb>0.0)
      selectionMask = (detecNullProb>minHistoX)
    end if

    numUseableSources = count(selectionMask)
if (numUseableSources <= 0) call error('noUseableSourcesInSet',"")

    allocate(log10DetecNullProb(numUseableSources))
    log10DetecNullProb = log10(pack(detecNullProb, selectionMask))

    call calcBinEdges(histoBinSize, log10DetecNullProb, binEdges)
    ! in ssclib/histogram_utils.

    numHistoBins = size(binEdges)-1

!    if (withSimSources) then
      ! Make histo of all valid detected sources:
      !
      allocate(detTotalHisto(numHistoBins), dummyWeights(numUseableSources))
      dummyWeights = 1.0
      call makeWeightedHistogram(log10DetecNullProb, dummyWeights, binEdges&
        , detTotalHisto, dummyUnderflows, dummyOverflows) ! in ssclib/histogram_utils.
      ! I have used the weighted routine just to get the xLo and xHi fields of the
      ! weightedHistoSingleType structure.
      deallocate(log10DetecNullProb, dummyWeights)

    if (withSimSources) then
      ! Now make histo of the subset of valid detected sources which are
      ! thought to be spurious detections:
      !
      selectionMask = (simInverseSensy>0.0 .and. detecNullProb>0.0&
        .and. matchNullProb > nullProbCutoff)
      numUnreliableSources = count(selectionMask)
      numReliableSources = numUseableSources - numUnreliableSources

      if (numReliableSources <= 0) then
        write(errStr, '(a, f6.4, a, e9.5, a)') "There are no detected sources &
          &with MATCH_PNULL less than the cutoff value ", nullProbCutoff&
          , ". Smallest MATCH_PNULL value found was ", minval(matchNullProb), "."
call error('noGoodSources')

      else if (numReliableSources == 1) then
        write(errStr, '(a, f6.4, a)') "Only one detected source was found with&
          & MATCH_PNULL less than the cutoff value ", nullProbCutoff, "."
call error('tooFewGoodSources')
      end if

    else ! no sim sources
      numUnreliableSources = numUseableSources
!      deallocate(log10DetecNullProb)
    end if

    allocate(detBadHistoInt(numHistoBins))

    if (numUnreliableSources <= 0) then
      detBadHistoInt = 0
    else
      allocate(log10DetecNullProb(numUnreliableSources))
      log10DetecNullProb = log10(pack(detecNullProb, selectionMask))

      call makeHistogram(log10DetecNullProb, detBadHistoInt, binEdges&
        , dummyUnderflows, dummyOverflows)

      deallocate(log10DetecNullProb)
    end if

    histoSet = dataSet(histoSetName, CREATE)
    histoTab = addTable(histoSet, 'HISTO', numHistoBins)

    call setAttribute(histoTab, 'SKY_AREA'&
      , real64Attribute(detSrcTab, 'SKY_AREA'), 'deg^2')
    call setAttribute(histoTab, 'N_FIELDS'&
      , int32Attribute(detSrcTab, 'N_FIELDS'))

    nBadCol              = addColumn(histoTab, 'N_BAD',      REAL32)
    nBadErrCol           = addColumn(histoTab, 'N_BAD_ERR',  REAL32)
    nBadIntCol           = addColumn(histoTab, 'N_BAD_INT',  REAL32)
    detectNullProbLoCol  = addColumn(histoTab, 'DPNUL_INT',  REAL32)
    detectNullProbMidCol = addColumn(histoTab, 'DPNUL_MID',  REAL32)
    detectNullProbHiCol  = addColumn(histoTab, 'DPNUL_HI',   REAL32)

    nBad              => real32Data(nBadCol)
    nBadErr           => real32Data(nBadErrCol)
    nBadInt           => real32Data(nBadIntCol)
    detectNullProbLo  => real32Data(detectNullProbLoCol)
    detectNullProbMid => real32Data(detectNullProbMidCol)
    detectNullProbHi  => real32Data(detectNullProbHiCol)

    nBad = detBadHistoInt / (1.0 - nullProbCutoff) ! the correction is
    ! done because the false detections are spread evenly at all values
    ! of MATCH_PNULL - ie some spurious detections will actually by
    ! chance have low values of MATCH_PNULL, and so will not be counted
    ! into the initial histogram.

    nBadErr = sqrt(nBad) / (1.0 - nullProbCutoff)
    nBadInt(0) = nBad(0)
    do bin = 1, numHistoBins-1
      nBadInt(bin) = nBadInt(bin-1) + nBad(bin)
    end do

    detectNullProbLo  = 10.0**detTotalHisto%xLo
    detectNullProbHi  = 10.0**detTotalHisto%xHi
    detectNullProbMid = 10.0**((detTotalHisto%xLo + detTotalHisto%xHi) / 2.0)

    if (withSimSources) then
      nAllCol     = addColumn(histoTab, 'N_ALL',      INTEGER32)
      nAllErrCol  = addColumn(histoTab, 'N_ALL_ERR',  REAL32)
      nAllIntCol  = addColumn(histoTab, 'N_ALL_INT',  INTEGER32)
      nTruCol     = addColumn(histoTab, 'N_TRUE',     REAL32)
      nTruErrCol  = addColumn(histoTab, 'N_TRUE_ERR', REAL32)
      nTruIntCol  = addColumn(histoTab, 'N_TRUE_INT', REAL32)

      nAll     =>  int32Data(nAllCol)
      nAllErr  => real32Data(nAllErrCol)
      nAllInt  =>  int32Data(nAllIntCol)
      nTru     => real32Data(nTruCol)
      nTruErr  => real32Data(nTruErrCol)
      nTruInt  => real32Data(nTruIntCol)

      nAll = detTotalHisto%n
      nAllErr = sqrt(real(nAll))
      nAllInt(0) = nAll(0)
      do bin = 1, numHistoBins-1
        nAllInt(bin) = nAllInt(bin-1) + nAll(bin)
      end do

      nTru = nAll - nBad
      nTruErr = sqrt(real(nTru) + nBadErr * nBadErr)
      nTruInt = nAllInt - nBadInt

      call release(nTruIntCol)
      call release(nTruErrCol)
      call release(nTruCol)
      call release(nAllIntCol)
      call release(nAllErrCol)
      call release(nAllCol)
    end if ! withSimSources

    call release(detectNullProbHiCol)
    call release(detectNullProbMidCol)
    call release(detectNullProbLoCol)
    call release(nBadIntCol)
    call release(nBadErrCol)
    call release(nBadCol)

    call release(histoTab)
    call release(histoSet)

    if (withSimSources) then
      call release(matchNullProbCol)
      call release(simInverseSensyCol)
    end if

    call release(detecNullProbCol)
    call release(detSrcTab)
    call release(detSrcSet)

    deallocate(binEdges)

  end subroutine calcReliabilityHistogramPart1

  !---------------------------------------------------------------------
  subroutine calcReliabilityHistogramPart2
  !---------------------------------------------------------------------
  ! Makes a single fits file that contains some histograms (usually over
  ! DET_ML). The column names are as follows:
  !
  ! Pre-existing:
  !
  !	N_ALL		Number sources detected.
  !	N_ALL_ERR	Square root of N_ALL.
  !	N_ALL_INT	Integrated value of N_ALL.
  !	N_BAD		Number of unreliable detections.
  !	N_BAD_ERR	Square root of N_BAD.
  !	N_BAD_INT	Integrated value of N_BAD.
  !	N_TRUE		Number of reliable detections.
  !	N_TRUE_ERR	Square root of N_TRUE.
  !	N_TRUE_INT	Integrated value of N_TRUE.
  !	DPNUL_INT	Lower edge of the bin.
  !	DPNUL_MID	Middle of the bin.
  !	DPNUL_HI	Upper edge of the bin.
!  !	DENS_ALL	Average sky-density (in deg^-2) of all detections.
!  !	DENS_ALL_ERR	Error in DENS_ALL.
!  !	DENS_ALL_INT	Integrated total of DENS_ALL.
!  !	DENS_TRUE	Average sky-density (in deg^-2) of reliable detections.
!  !	DENS_TRUE_ERR	Error in DENS_TRUE.
!  !	DENS_TRUE_INT	Integrated total of DENS_TRUE.
  !
  ! New:
  !
  !	FIRMFRAC	DENS_TRUE/DENS_ALL.
  !	FIRMFRAC_ERR	Error in FIRMFRAC.
  !	FIRMFRAC_INT	DENS_TRUE_INT/DENS_ALL_INT.

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    integer :: numHistoBins, bin
    type(ColumnT) ::&
      nAllCol, nAllErrCol, nAllIntCol, nTruCol, nTruErrCol, nTruIntCol,&
      firmFractionCol, firmFractionIntCol, firmFractionErrCol
    integer(int32), pointer :: nAll(:), nAllInt(:)
    real(single), pointer ::&
      nAllErr(:), nTru(:), nTruErr(:), nTruInt(:), firmFraction(:),&
      firmFractionInt(:), firmFractionErr(:)

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making FIRMFRAC histogram.")

    histoSet = dataSet(histoSetName, MODIFY)
    histoTab = table(histoSet, 'HISTO')
    numHistoBins = numberOfRows(histoTab)

    if (hasColumn(histoTab, 'N_ALL')) then
      nAllCol     = column(histoTab, 'N_ALL')
      nAllErrCol  = column(histoTab, 'N_ALL_ERR')
      nAllIntCol  = column(histoTab, 'N_ALL_INT')
      nTruCol     = column(histoTab, 'N_TRUE')
      nTruErrCol  = column(histoTab, 'N_TRUE_ERR')
      nTruIntCol  = column(histoTab, 'N_TRUE_INT')
      firmFractionCol    = addOrOpenColumn(histoTab, 'FIRMFRAC',     REAL32)
      firmFractionIntCol = addOrOpenColumn(histoTab, 'FIRMFRAC_INT', REAL32)
      firmFractionErrCol = addOrOpenColumn(histoTab, 'FIRMFRAC_ERR', REAL32)
      nAll     =>  int32Data(nAllCol)
      nAllErr  => real32Data(nAllErrCol)
      nAllInt  =>  int32Data(nAllIntCol)
      nTru     => real32Data(nTruCol)
      nTruErr  => real32Data(nTruErrCol)
      nTruInt  => real32Data(nTruIntCol)

      do bin = 0, numHistoBins - 1
        if (nAll(bin) <= 0.0) call setToNull(firmFractionCol, bin)
      end do

      do bin = 0, numHistoBins - 1
        if (nAllInt(bin) <= 0.0) call setToNull(firmFractionIntCol, bin)
      end do

      do bin = 0, numHistoBins - 1
        if (nAllErr(bin) <= 0.0 .or. nTruErr(bin) <= 0.0)&
          call setToNull(firmFractionErrCol, bin)
      end do

      firmFraction    => real32Data(firmFractionCol)
      firmFractionInt => real32Data(firmFractionIntCol)
      firmFractionErr => real32Data(firmFractionErrCol)

      where(nAll > 0.0)&
        firmFraction = min(1.0, max(0.0, nTru / real(nAll, single)))

      where(nAllInt > 0.0)&
        firmFractionInt = min(1.0, max(0.0, nTruInt / real(nAllInt, single)))

      where(nAllErr > 0.0 .and. nTruErr > 0.0)
        firmFractionErr = firmFraction&
        *sqrt(nAllErr * nAllErr / real(nAll, single) / real(nAll, single)&
        +     nTruErr * nTruErr / real(nTru, single) / real(nTru, single))
      end where

      call release(firmFractionErrCol)
      call release(firmFractionIntCol)
      call release(firmFractionCol)
      call release(nTruIntCol)
      call release(nTruErrCol)
      call release(nTruCol)
      call release(nAllIntCol)
      call release(nAllErrCol)
      call release(nAllCol)
    end if

    call release(histoTab)
    call release(histoSet)

  end subroutine calcReliabilityHistogramPart2

  !---------------------------------------------------------------------
  subroutine makePlot1
  !---------------------------------------------------------------------
  ! Make a scatter plot of x=DPNUL_MID, y=N_BAD/N_FIELDS, y_err=N_BAD_ERR/N_FIELDS; both logarithmic axes. The number of false detections N_BAD should be proportional to the null probability DPNUL, thus this log-log plot should have slope of 1. The proportionality is the number of 'beams' in the image.

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    real(single), pointer :: xVals(:), yVals(:), yErrs(:)
    integer :: numRows, firstNonEmptyBin, numFields
    character(300) :: pgplotDevice

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 1.")

    histoSet = dataSet(histoSetName, READ)
    histoTab = table(histoSet, 'HISTO')

    numFields = int32Attribute(histoTab, 'N_FIELDS')

    numRows = numberOfRows(histoTab)
    if (numRows <= 0) then
      call warning('noRows',"")
      call release(histoTab)
      call release(histoSet)
  return
    end if

    call readColDataToPtr(histoTab, 'DPNUL_MID', xVals)
    call readColDataToPtr(histoTab, 'N_BAD',     yVals)
    call readColDataToPtr(histoTab, 'N_BAD_ERR', yErrs)
    ! In ssclib/dal_aux.

    yVals = yVals / real(numFields, single)
    yErrs = yErrs / real(numFields, single)

    ! No use plotting histogram bins until we start finding some bad sources within them. So find the first such bin:
    !
    firstNonEmptyBin = minloc1d(xVals, (yVals > 0.0)) ! in ssclib/array_utils.
    if (firstNonEmptyBin > numRows) then
      call warning('allRowsEmpty',"")
    else
      write (pgplotDevice, '(a,a)') trim(plot1FileName), pgDev

      call plotInit(pgplotDevice, 'DETEC_PNULL', 'Number of false detections per field')
      call setXLog
      call setYLog
      call drawPointPlot(xVals(firstNonEmptyBin-1:), yVals(firstNonEmptyBin-1:), yErrs(firstNonEmptyBin-1:))
      ! All in ssclib/pgplot_aux.
    end if

    deallocate(yErrs, yVals, xVals)

    call release(histoTab)
    call release(histoSet)

  end subroutine makePlot1

  !---------------------------------------------------------------------
  subroutine makePlot2
  !---------------------------------------------------------------------
  ! Same as plot1, but a line plot from the cumulative integral of N_BAD. Here x=DPNUL_INT and y=N_BAD_INT; both logarithmic axes.

    implicit none

    type(DataSetT) :: histoSet
    type(TableT) :: histoTab
    integer :: numRows, firstNonEmptyBin, numFields
    real(single), pointer :: xVals(:), yVals(:)
    character(300) :: pgplotDevice

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 2.")

    histoSet = dataSet(histoSetName, READ)
    histoTab = table(histoSet, 'HISTO')

    numFields = int32Attribute(histoTab, 'N_FIELDS')

    numRows = numberOfRows(histoTab)
    if (numRows <= 0) then
      call warning('noRows',"")
      call release(histoTab)
      call release(histoSet)
  return
    end if

    call readColDataToPtr(histoTab, 'DPNUL_INT', xVals)
    call readColDataToPtr(histoTab, 'N_BAD_INT', yVals)
    ! In ssclib/dal_aux.

    yVals = yVals / real(numFields, single)

    ! No use plotting histogram bins until we start finding some bad sources within them. So find the first such bin:
    !
    firstNonEmptyBin = minloc1d(xVals, (yVals > 0.0)) ! in ssclib/array_utils.
!    lastNonEmptyBin = maxloc1d(xVals, (yVals > 0.0)) ! in ssclib/array_utils.
!    lastNonEmptyBin = maxLoc1dSingle(xVals, (yVals > 0.0)) ! in ssclib/array_utils.
!print *, 'firstNonEmptyBin, lastNonEmptyBin=', firstNonEmptyBin, lastNonEmptyBin
!print *, yVals(firstNonEmptyBin-1:)
    if (firstNonEmptyBin > numRows .or. firstNonEmptyBin < 1) then
      call warning('allRowsEmpty',"")
    else if (minval(yVals(firstNonEmptyBin:)) == maxval(yVals(firstNonEmptyBin:))) then
      call warning('onlyOnePoint',"") ! Can't plot only 1 point, at least not using drawLinePlot.
    else
      write (pgplotDevice, '(a,a)') trim(plot2FileName), pgDev

      call plotInit(pgplotDevice, 'DETEC_PNULL', 'Cumulative number of false detections per field')
      call setXLog
      call setYLog
      call drawLinePlot(xVals(firstNonEmptyBin-1:), yVals(firstNonEmptyBin-1:))
    ! All in ssclib/pgplot_aux.
    end if

    deallocate(yVals, xVals)

    call release(histoTab)
    call release(histoSet)

  end subroutine makePlot2

  !---------------------------------------------------------------------
  subroutine makePlot3
  !---------------------------------------------------------------------
  ! Plot 3 is a simple comparison between the fluxes of detected and simulated sources. Three plots are superimposed as follows:
  !	1:	x=SIM_FLUX	y=FLUX	y_err=FLUX_ERR		red point+err
  !	2:	A straight line of slope 1			blue line
  !	3:	x=SIM_FLUX	y=FLUX	y_err=none		black point

    implicit none

    type(DataSetT) :: detSrcSet
    type(TableT) :: detSrcTab
    real(single), pointer :: xVals(:), yVals(:), yErrs(:)
    character(300) :: pgplotDevice
    real(single) :: sLo, sHi
    logical :: withSimSources
    logical, allocatable :: pointsNotLines(:)
    integer, allocatable :: styles(:), colours(:)
    character(30), allocatable :: texts(:)
    integer :: num

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 3.")

    detSrcSet = dataSet(detSrcSetName, READ)
    detSrcTab = table(detSrcSet, 'SRCLIST')

    withSimSources = .false. ! default
    if (hasAttribute(detSrcTab, 'COMPARED')) then
      if (booleanAttribute(detSrcTab, 'COMPARED')) withSimSources = .true.
    end if

    if (withSimSources) then
      call readColDataToPtr(detSrcTab, 'SIM_FLUX', xVals)
      call readColDataToPtr(detSrcTab, 'FLUX',     yVals)
      call readColDataToPtr(detSrcTab, 'FLUX_ERR', yErrs)
      ! In ssclib/dal_aux.

      sLo = min(minval(xVals), minval(yVals)) / 10.0
      sHi = max(maxval(xVals), maxval(yVals)) * 10.0

num = 3
allocate(pointsNotLines(num), styles(num), colours(num), texts(num))
pointsNotLines = (/.true.,.true.,.false./)
styles = (/P_DEFAULT, P_DEFAULT, L_DEFAULT/)
colours = (/1,2,3/)
texts = (/'Detections                '&
         ,'Same, but with flux errors'&
         ,'Theory                    '/)

      write (pgplotDevice, '(a,a)') trim(plot3FileName), pgDev

      call plotInit(pgplotDevice, 'SIM_FLUX (cgs)', 'Detected FLUX (cgs)')
      call setXLog
      call setYLog
      call drawPointPlot(xVals, yVals, yErrs,     colourIndex=2, finished=.false.)
      call drawLinePlot((/sLo,sHi/), (/sLo,sHi/), colourIndex=3, finished=.false.)
      call drawPointPlot(xVals, yVals,            colourIndex=1, finished=.false.)
call drawLegend(pointsNotLines, styles, colours, texts, position=XY_TOPLEFT)
call finishPlot
    ! All in ssclib/pgplot_aux.
deallocate(pointsNotLines, styles, colours, texts)

      deallocate(yErrs, yVals, xVals)
    else
      call warning('noSimSrcData',"")
    end if

    call release(detSrcTab)
    call release(detSrcSet)

  end subroutine makePlot3

  !---------------------------------------------------------------------
  subroutine makePlot4
  !---------------------------------------------------------------------
  ! Plot 4 shows how the 'match' probability varies with flux.
  !	x=SIM_FLUX	y=MATCH_PNULL	point plot, X axis logarithmic, Y linear.

    implicit none

    type(DataSetT) :: detSrcSet
    type(TableT) :: detSrcTab
    real(single), pointer :: xVals(:), yVals(:)
    character(300) :: pgplotDevice
    logical :: withSimSources
    real(single) :: xLo, xHi

    !-------------------------------------------------------------------
    ! Start

    call message(AppMsg, Verbose, "Making plot 4.")

    detSrcSet = dataSet(detSrcSetName, READ)
    detSrcTab = table(detSrcSet, 'SRCLIST')

    withSimSources = .false. ! default
    if (hasAttribute(detSrcTab, 'COMPARED')) then
      if (booleanAttribute(detSrcTab, 'COMPARED')) withSimSources = .true.
    end if

    if (withSimSources) then
      call readColDataToPtr(detSrcTab, 'SIM_FLUX',    xVals)
      call readColDataToPtr(detSrcTab, 'MATCH_PNULL', yVals)
      ! In ssclib/dal_aux.

      xLo = minval(xVals) / 10.0
      xHi = maxval(xVals) * 10.0

      write (pgplotDevice, '(a,a)') trim(plot4FileName), pgDev

      call plotInit(pgplotDevice, 'SIM_FLUX (cgs)', 'MATCH_PNULL')
      call setXLog
!      call setYLog
      call plotAutoSizeX(xVals)
      call plotAutoSizeY(yVals)
      call drawLinePlot((/xLo,xHi/), (/nullProbCutoff,nullProbCutoff/)&
        , lineStyle=L_DASHED, finished=.false.)
      call drawPointPlot(xVals, yVals, pointStyle=P_DIAMOND)
      ! All in ssclib/pgplot_aux.

      deallocate(yVals, xVals)
    else
      call warning('noSimSrcData',"")
    end if

    call release(detSrcTab)
    call release(detSrcSet)

  end subroutine makePlot4

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module eimsimreliability_mod
