! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
! Module: power_law_utils
!
! Authors:  Ian Stewart - SSC/LUX
!          
!#######################################################################
! Description: Contains routines to support logN-logS curves made up of
! piecewise power-laws.
!
! Suppose we have an ensemble of sources with fluxes S. Let us consider
! n(S)dS, the density of sources per unit sky area and per unit flux
! interval as a function of the source flux. Of interest is the reverse
! cumulative sky density N(>S), which is defined as
!
!	       inf
!	        /
!	N(>S) = | dS n(S).
!	        /
!	       S
!
! The present module is solely concerned with N(>S) in the form of
! a sequence of power laws, that is,
!
!	N(>S) = K_i S^{G_i}, S_i <= S < S_{i+1}, for i in [1,m],
!
! given a sequence of 'anchor' or 'knee' points S_i. (S_(m+1) is supposed to be at infinity.) For this to be
! possible implies that n(S) must also be a piecewise power law
!
!		n(S)dS = k_i S^{g_i} dS, S_i <= S < S_{i+1}, for i in [1,m],	(1)
!
! where G_i = g_i + 1 and
!
!	        -k_i
!	K_i = ---------.
!	       1 + g_i
!
! Note the minus sign; note also that physicality requires that (i) N(>S) be continuous; (ii) n = 0 for S < S_1; (iii) g_m < -1.
!#######################################################################
! 
! Comments: TODO:
! 
! $Id: power_law_utils.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module power_law_utils

  use dal
  use errorHandling

  use array_utils
!use temp_ssclib
!use l_array_utils_2
  use dal_aux	! for readColDataToPtr

  implicit none 

  private 

  public getLogNLogS
  public invertLogNlogSrandoms
  public ampErrToFluxRandErr
  private doLogNlogSchecks
  public calcLogNLogSPowers
  public calcKVals
  public calcAccumFluxCoeffs
  public calcReverseAccumFluxCoeffs
  public calcNumDensity
  public calcCumFluxDensity
  public calcCumFluxDensities
  public calcReverseCumFluxDensity
  public calcReverseCumNumDensity

contains 

  !---------------------------------------------------------------------
  subroutine getLogNlogS(logNlogStab, modelFluxes, modelCumSkyDensity&
  , hiPower)
  !---------------------------------------------------------------------
  ! A cumulative function N(>S) which is a series of power laws as described in the header has the property that, when plotted in log-log format, ie when log(N) is plotted against log(S), the function appears as a connected series of straight lines. Such a function is thus easily parameterized by the coordinates of all the 'knee' points, plus the slope of the last, high-S section. This is the way it is stored in the logNlogStab FITS table.
  !
  ! Remember to deallocate pointers modelFluxes(:) and
  ! modelCumSkyDensity(:) after use.

    implicit none

    type(TableT), intent(in)  :: logNlogStab
    real(single), pointer     :: modelFluxes(:),&
                                 modelCumSkyDensity(:)
    real(single), intent(out) :: hiPower

    integer(int32) :: numFluxes

    !-------------------------------------------------------------------
    ! Start

    numFluxes = numberOfRows(logNlogStab)
if (numFluxes < 1) call error('noFluxValues','')
    call readColDataToPtr(logNlogStab, 'FLUX',    modelFluxes)
    call readColDataToPtr(logNlogStab, 'DENSITY', modelCumSkyDensity)
    ! readColDataToPtr is in ssclib/dal_aux.

    if (numFluxes > 1) then
      if (any(modelCumSkyDensity(:numFluxes-2) < modelCumSkyDensity(1:)))&
call error('badDensityOrder','')
      if (any(modelFluxes(:numFluxes-2) >= modelFluxes(1:)))&
call error('badFluxOrder','')
    end if
if (modelFluxes(0) <= 0.0) call error('nonPositiveFlux','')
if (modelCumSkyDensity(0) <= 0.0) call error('nonPositiveSkyDensity','')

    if (hasAttribute(logNlogStab, 'HI_SLOPE')) then
      hiPower = real32Attribute(logNlogStab, 'HI_SLOPE')
if (hiPower >= 0.0) call error('nonNegativeHiPower',"")
    else
      hiPower = -1.5
    end if

  end subroutine getLogNlogS

  !---------------------------------------------------------------------
  subroutine getLogNlogS_20060509(logNlogStab, modelFluxDensity0, modelCumSkyDensity&
  , hiPower)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)            :: logNlogStab
    real(single), pointer               :: modelFluxDensity0(:),&
                                           modelCumSkyDensity(:)
    real(single), intent(out), optional :: hiPower

    integer(int32) :: numFluxes

    !-------------------------------------------------------------------
    ! Start

    numFluxes = numberOfRows(logNlogStab)
if (numFluxes < 1) call error('noFluxValues','')
    call readColDataToPtr(logNlogStab, 'S_DENS_0',  modelFluxDensity0)
    call readColDataToPtr(logNlogStab, 'SKY_DENSY', modelCumSkyDensity)
    ! readColDataToPtr is in ssclib/dal_aux.

    if (numFluxes > 1) then
      if (any(modelCumSkyDensity(:numFluxes-2) < modelCumSkyDensity(1:)))&
call error('badDensityOrder','')
      if (any(modelFluxDensity0(:numFluxes-2) >= modelFluxDensity0(1:)))&
call error('badFluxOrder','')
    end if
if (modelFluxDensity0(0) <= 0.0) call error('nonPositiveFluxDensity','')
if (modelCumSkyDensity(0) <= 0.0) call error('nonPositiveSkyDensity','')

    if (present(hiPower)) hiPower = real32Attribute(logNlogStab, 'HI_SLOPE')

  end subroutine getLogNlogS_20060509

  !---------------------------------------------------------------------
  subroutine invertLogNlogSrandoms(x, yInt, xSamples, hiPower, harvest&
  , status)
  !---------------------------------------------------------------------
  ! This sub was modified from the routine getLogNlogSrandoms in random.f90.
  !.....................................................................
  !
  ! The subroutine takes a vector of random values xSamples() which have a
  ! distribution P which, when integrated, follows a piecewise power law
  ! as follows. The ith segment of the integrated piecewise function is
  ! defined between x(i) and x(i+1) and is of the form
  !
  !	log(yInt) = log(yInt(i)) + slope * (log(x) - log(x(i)))
  !
  ! where
  !	         log(yInt(i+1)) - log(yInt(i))
  !	slope = -------------------------------.
  !	            log(x(i+1)) - log(x(i))
  !
  ! Above x(N) the integrated function is defined by
  !
  !	log(yInt) = log(yInt(N)) + hiPower * (log(x) - log(x(N))).
  !
  ! In the present subroutine the original random values, which are evenly
  ! distributed between 0 and 1, are returned.
  !.....................................................................
  !
  ! Notes: (i) The x values must be in increasing order, and they cannot be
  ! negative. (ii) The y values must all be >0, and must never increase with
  ! increasing i. (iii) N must be >0. (iv) hiPower must be <0.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................

    implicit none

    real,    intent(in)            :: x(:),&
                                      yInt(size(x)),&
                                      xSamples(:),&
                                      hiPower
    real,    intent(out)           :: harvest(size(xSamples))
    integer, intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numBins, numXsamples, i, bin
    real :: slopes(size(x))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    harvest = -1

    numBins = size(x)
!print *, lbound(x), ubound(x)
    numXsamples = size(xSamples)

    call doLogNlogSchecks(x, yInt, hiPower, foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    ! Calculate slopes:

    do bin = 1, numBins - 1
      slopes(bin) = (log(yInt(bin+1)) - log(yInt(bin)))&
                  / (log(x(   bin+1)) - log(x(   bin)))
    end do
    slopes(numBins) = hiPower

    ! Main part of the routine:

    do i = 1, numXsamples
      bin = getBinNumber(x, xSamples(i)) ! in ssclib/array_utils
if (bin < 1) bin = 1 ! to cope with zero- and non-(ID_BAND==0)-RATE values.
      harvest(i) = yInt(bin) * (xSamples(i) / x(bin))**slopes(bin)
    end do
    harvest = harvest / yInt(1)

    if (present(status)) status = localStatus

  end subroutine invertLogNlogSrandoms

  !---------------------------------------------------------------------
  subroutine ampErrToFluxRandErr(x, yInt, xSamples, xErrs, hiPower, randErrs&
  , status)
  !---------------------------------------------------------------------
  ! This sub was modified from invertLogNlogSrandoms.
  !.....................................................................
  !
  ! The subroutine takes a vector of random values xSamples() which have a
  ! distribution P which, when integrated, follows a piecewise power law
  ! as follows. The ith segment of the integrated piecewise function is
  ! defined between x(i) and x(i+1) and is of the form
  !
  !	log(yInt) = log(yInt(i)) + slope * (log(x) - log(x(i)))
  !
  ! where
  !	         log(yInt(i+1)) - log(yInt(i))
  !	slope = -------------------------------.
  !	            log(x(i+1)) - log(x(i))
  !
  ! Above x(N) the integrated function is defined by
  !
  !	log(yInt) = log(yInt(N)) + hiPower * (log(x) - log(x(N))).
  !
  ! In the present subroutine the original random values, which are evenly
  ! distributed between 0 and 1, are returned.
  !.....................................................................
  !
  ! Notes: (i) The x values must be in increasing order, and they cannot be
  ! negative. (ii) The y values must all be >0, and must never increase with
  ! increasing i. (iii) N must be >0. (iv) hiPower must be <0.
  !
  ! In the event of an error, the subroutine returns values of -1 for all
  ! elements of numbers().
  !.....................................................................

    implicit none

    real,    intent(in)            :: x(:),&
                                      yInt(size(x)),&
                                      xSamples(:),&
                                      xErrs(size(xSamples)),&
                                      hiPower
    real,    intent(out)           :: randErrs(size(xSamples))
    integer, intent(out), optional :: status

    integer ::&
      localStatus, foreignStatus, numBins, numXsamples, i, bin
    real :: slopes(size(x))

    !-------------------------------------------------------------------
    ! Start

    localStatus = 0
    randErrs = -1

    numBins = size(x)
    numXsamples = size(xSamples)

    call doLogNlogSchecks(x, yInt, hiPower, foreignStatus)

    if (foreignStatus < 0) then
      if (present(status)) status = foreignStatus
  return
    end if

    ! Calculate slopes:

    do bin = 1, numBins - 1
      slopes(bin) = (log(yInt(bin+1)) - log(yInt(bin)))&
                  / (log(x(   bin+1)) - log(x(   bin)))
    end do
    slopes(numBins) = hiPower

    ! Main part of the routine:

    do i = 1, numXsamples
      bin = getBinNumber(x, xSamples(i)) ! in array_utils
!      bin = l_getBinNumberSingle(x, xSamples(i)) ! in ssclib/array_utils
if (bin < 1) bin = 1 ! to cope with zero- and non-(ID_BAND==0)-RATE values.
      randErrs(i) = xErrs(i) * slopes(bin) * yInt(bin) / x(bin)&
        * (xSamples(i) / x(bin))**(slopes(bin) - 1.0)
    end do
    randErrs = abs(randErrs / yInt(1))

    if (present(status)) status = localStatus

  end subroutine ampErrToFluxRandErr

  !---------------------------------------------------------------------
  subroutine doLogNlogSchecks(x, yInt, hiPower, status)
  !---------------------------------------------------------------------
  ! Taken straight from ssclib/random_aux.f90.

    implicit none

    real,    intent(in)  :: x(:), yInt(size(x)), hiPower
    integer, intent(out) :: status

    integer :: numPoints

    !-------------------------------------------------------------------
    ! Start:

    status = 0
    numPoints = size(x)

    if (numPoints < 1) then
      status = -1
  return
    end if

    if (numPoints > 1) then ! Check order of x values:
      if (any(x(2:) <= x(:numPoints-1))) then
        status = -2
  return
      end if

      if (any(yInt(2:) > yInt(:numPoints-1))) then
        status = -3
  return
      end if
    end if

    ! Check all x values >= 0:
    if (x(1) < 0.0) then
      status = -4
  return
    end if

    ! Check all yInt values > 0:
    if (any(yInt <= 0.0)) then
      status = -5
  return
    end if

!    if (hiPower >= -1.0) then
if (hiPower >= 0.0) then
      status = -6
  return
    end if

  end subroutine doLogNlogSchecks

  !---------------------------------------------------------------------
  subroutine calcLogNLogSPowers(fluxes, cumSkyDensity, hiPower, powers)
  !---------------------------------------------------------------------
  ! The distribution of source fluxes for the ensemble of simulated sources is specified by a logN-logS curve (ie cumulative source density per square degree, as a function of flux S) which is made up of a series of connected line segments between 'anchor' or 'knee' points (log(S_i),log(N_i)). This translates to differential source density n(S)dS per square degree per unit flux which is a sequence of power laws
  !
  !	n(S)dS = k_i S^{g_i} dS.
  !
  ! The present subroutine takes the set of Ns and Ss and returns a vector which contains the equivalent values of g_i.
  !
  !---------------------------------------------------------------------
  ! Tests out ok. IMS 2007-08-14.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: fluxes(:),&
                                 cumSkyDensity(size(fluxes)),&
                                 hiPower
    real(single), intent(out) :: powers(size(fluxes))

    integer :: numSamples, sampleNum

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(fluxes)
    do sampleNum = 1, numSamples - 1
      powers(sampleNum)&
        = (log(cumSkyDensity(sampleNum+1)) - log(cumSkyDensity(sampleNum)))&
        / (log(fluxes(       sampleNum+1)) - log(fluxes(       sampleNum)))
    end do

    powers(numSamples) = hiPower

    powers = powers - 1 ! to return the differential rather than integral
    ! powers.

  end subroutine calcLogNLogSPowers

  !---------------------------------------------------------------------
  subroutine calcKVals(fluxes, cumSkyDensity, powers, kVals)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: fluxes(:),&
                                 cumSkyDensity(size(fluxes)),&
                                 powers(size(fluxes))
    real(single), intent(out) :: kVals(size(fluxes))

    !-------------------------------------------------------------------
    ! Start

    kVals = -cumSkyDensity * (1.0 + powers) / (fluxes**(1.0+powers))

  end subroutine calcKVals

  !---------------------------------------------------------------------
  subroutine calcAccumFluxCoeffs(fluxes, kVals, powers, offsets, multipliers)
  !---------------------------------------------------------------------
  ! See the description in the module header of the sequential-power-law cumulative density function N(>S) which is the subject of all these routines. 
  !
  ! In the present program we want to calculate, not the number of sources per square degree, but the average flux per square degree. The differential average flux density f(S) is obtained from equation 1 in the header by multiplying by S, giving:
  !
  !	f(S)dS = k_i S^{1+g_i} dS.
  !
  ! Unfortunately, unlike N(>S), neither the cumulative integral F(<S) nor the inverse cumulative integral F(>S) are expressible as joined-up power laws at each value of S. We are interested here in the cumulative integral
  !
  !	         __j-1   S_{i+1}                  S
  !	        \          /                      /
  !	F(<S) =  >     k_i | dS S^{1+g_i}  +  k_j | dS S^{1+g_j};	(2)
  !	        /__        /                      /
  !	        i=1       S_i                    S_j
  !
  ! or
  !	        __j-1
  !	       \      k_i                                         k_j
  !	     =  >   ------- [S_{i+1}^{2+g_i} - S_i^{2+g_i}]  -  ------- S_j^{2+g_j}
  !	       /__   2+g_i                                       2+g_j
  !	       i=1
  !
  !	         k_j
  !	     + ------- S^{2+g_j},					(3)
  !             2+g_j
  !
  ! for S_j <= S < S_{j+1} (or < inf if j==m). Rather than calculate the whole of equation 3 for each new S, it is quicker to pre-calculate the first line of equation 3 (plus the premultiplier of the last line) for each j. That is the function of the present subroutine.
  !
  !---------------------------------------------------------------------
  ! Tests out ok. IMS 2007-08-14.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in)  :: fluxes(:),&
                                 kVals(size(fluxes)),&
                                 powers(size(fluxes))
    real(single), intent(out) :: offsets(size(fluxes)),&
                                 multipliers(size(fluxes))

    integer :: numSamples, i

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(fluxes)

    if (powers(numSamples) >= -2.0)&
call error('fluxIntegralUnbounded','')
    if (any(2.0 + powers == 0.0))& ! needs ln solution.
call error('notYetSupported','')

    multipliers = kVals / (2.0 + powers)
    offsets(1) = 0.0
    do i = 2, numSamples
      offsets(i) = offsets(i-1)&
        + multipliers(i-1)&
        * (fluxes(i)**(2.0 + powers(i-1)) - fluxes(i-1)**(2.0 + powers(i-1)))
    end do

    offsets = offsets - multipliers * fluxes**(2.0 + powers)

  end subroutine calcAccumFluxCoeffs

  !---------------------------------------------------------------------
  subroutine calcReverseAccumFluxCoeffs(fluxes, kVals, powers, offsets, multipliers)
  !---------------------------------------------------------------------
!*** not yet tested
  ! See the description in the module header of the sequential-power-law cumulative density function N(>S) which is the subject of all these routines. 
  !
  ! In the present program we want to calculate, not the number of sources per square degree, but the average flux per square degree. The differential average flux density f(S) is obtained from equation 1 in the header by multiplying by S, giving:
  !
  !	f(S)dS = k_i S^{1+g_i} dS.
  !
  ! Unfortunately, unlike N(>S), neither the cumulative integral F(<S) nor the inverse cumulative integral F(>S) are expressible as joined-up power laws at each value of S. We are interested here in the reverse cumulative integral
  !
  !	           S_{j+1}              __N-1   S_{i+1}
  !	             /                 \          /
  !	F(>S) =  k_j | dS S^{1+g_j} +   >     k_i | dS S^{1+g_i} +
  !	             /                 /__        /
  !	             S                i=j+1     S_i
  !
  !	            inf
  !	             /
  !	      +  k_N | dS S^{1+g_N};					(2)
  !	             /
  !	            S_N
  ! or
  !	                                  __N-1
  !	          k_j                    \      k_i
  !	     =  ------- S_{j+1}^{2+g_j} + >   ------- [S_{i+1}^{2+g_i} - S_i^{2+g_i}] -
  !	         2+g_j                   /__   2+g_i
  !	                                i=j+1
  !
  !
  !	         k_N
  !	     - ------- S_N^{2+g_N}
  !	        2+g_N
  !
  !
  !	         k_j
  !	     - ------- S^{2+g_j},					(3)
  !             2+g_j
  !
  ! for S_j <= S < S_(j+1) (or < inf if j==max_j). Rather than calculate the whole of equation 3 for each new S, it is quicker to pre-calculate the first two lines of equation 3 (plus the premultiplier of the last line) for each j. That is the function of the present subroutine.

    implicit none

    real(single), intent(in)  :: fluxes(:),&
                                 kVals(size(fluxes)),&
                                 powers(size(fluxes))
    real(single), intent(out) :: offsets(size(fluxes)),&
                                 multipliers(size(fluxes))

    integer :: numSamples, i

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(fluxes)

    if (powers(numSamples) >= -2.0)&
call error('fluxIntegralUnbounded','')
    if (any(2.0 + powers == 0.0))& ! needs ln solution.
call error('notYetSupported','')

    multipliers = kVals / (2.0 + powers)
    offsets(numSamples) = -multipliers(numSamples)&
      * fluxes(numSamples)**(2.0 + powers(numSamples))
    do i = numSamples-1, 1, -1
      offsets(i) = offsets(i+1)&
        + multipliers(i)&
        * (fluxes(i+1)**(2.0 + powers(i)) - fluxes(i)**(2.0 + powers(i)))
    end do

    offsets(2:) = offsets(2:) + multipliers(1:numSamples-1)&
      * fluxes(2:)**(2.0 + powers(1:numSamples-1))
    multipliers = -multipliers

!do i = 1, numSamples
!  print *, i, fluxes(i), kVals(i), powers(i), offsets(i), multipliers(i)
!end do

  end subroutine calcReverseAccumFluxCoeffs

  !---------------------------------------------------------------------
  subroutine calcAccumFluxes_old(fluxes, cumSkyDensity, powers, offsets, multipliers)
  !---------------------------------------------------------------------
!*** not yet tested
  ! The distribution of source fluxes for the ensemble of simulated sources is specified by a logN-logS curve (ie cumulative source density per square degree, as a function of flux S) which is made up of a series of connected line segments between 'anchor' or 'knee' points (log(S_i),log(N_i)). This translates to differential source density n(S)dS per square degree per unit flux which is a sequence of power laws
  !
  !	n(S)dS = k_i S^(-g_i),						(1)
  !
  ! where the continuity of the integrated density at knee points (and the linearity of log(N(log(S))) between them) requires that
  !
  !	   k_i                         k_(i+1)
  !	--------- S_(i+1)^(1-g_i) = ------------- S_(i+1)^(1-g_(i+1))
  !	 1 - g_i                     1 - g_(i+1)
  !
  ! and thus also
  !
  !	   k_i
  !	--------- S_i^(1-g_i) = -N_i.
  !	 1 - g_i
  !
  ! In the present program we want to calculate, not the number of sources per square degree, but the average flux per square degree. The differential average flux density is obtained from equation 1 by multiplying by S:
  !
  !	f(S)dS = k_i S^(1-g_i).
  !
  ! Unfortunately the cumulative integral of this, F(S), is, unlike N(S), no longer expressible as a joined-up power law at each S. We are interested here in the integral
  !
  !	        __j-1   S_(i+1)                  S
  !	       \          /                      /
  !	F(S) =  >     k_i | dS S^(1-g_i)  +  k_j | dS S^(1-g_j);	(2)
  !	       /__        /                      /
  !	       i=1       S_i                    S_j
  !
  ! or
  !	        __j-1
  !	       \      k_i                                         k_j
  !	     =  >   ------- [S_(i+1)^(2-g_i) - S_i^(2-g_i)]  -  ------- S_j^(2-g_j)
  !	       /__   2-g_i                                       2-g_j
  !	       i=1
  !
  !	         k_j
  !	     + ------- S^(2-g_j),					(3)
  !             2-g_j
  !
  ! for S_j <= S < S_(j+1) (or < inf if j==max_j). Rather than calculate the whole of equation 2 for each new S, it is quicker to pre-calculate the first line of equation 3 (plus the premultiplier of the last line) for each j. That is the function of the present subroutine.

    implicit none

    real(single), intent(in)  :: fluxes(:),&
                                 cumSkyDensity(size(fluxes)),&
                                 powers(size(fluxes))
    real(single), intent(out) :: offsets(size(fluxes)),&
                                 multipliers(size(fluxes))

    integer :: numSamples, i
    real(single), allocatable :: kVals(:)

    !-------------------------------------------------------------------
    ! Start

    numSamples = size(fluxes)

    if (powers(numSamples) >= -2.0)&
call error('fluxIntegralUnbounded','')
    if (any(2.0 + powers == 0.0))& ! needs ln solution.
call error('notYetSupported','')

    allocate(kVals(numSamples))
    !	   k_i
    !	--------- S_i^(1-g_i) = N_i.
    !	 1 - g_i
    kVals = cumSkyDensity * (1.0 + powers) / (fluxes**(1.0+powers))
    multipliers = kVals / (2.0 + powers)
    offsets(1) = 0.0
    do i = 2, numSamples
      offsets(i) = offsets(i-1)&
        + multipliers(i-1)&
        * (fluxes(i)**(2.0 + powers(i-1)) - fluxes(i-1)**(2.0 + powers(i-1)))
    end do

    offsets = offsets - multipliers * fluxes(i)**(2.0 + powers)

  end subroutine calcAccumFluxes_old

  !---------------------------------------------------------------------
  function calcNumDensity(fluxSample, fluxes, kVals, powers) result(numDensity)
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: fluxSample,&
                                fluxes(:),&
                                kVals(size(fluxes)),&
                                powers(size(fluxes))
    real(single)             :: numDensity

    integer(int32) :: binNum

    !-------------------------------------------------------------------
    ! Start:

if (size(fluxes)<1) call error('bug',"")
    binNum = getBinNumber(fluxes, fluxSample) ! in ssclib/array_utils.
if (binNum <= 0) call error('bug',"")

    numDensity = kVals(binNum) * (fluxSample**powers(binNum))
!print *, 'binNum, numDensity, kVals(binNum), fluxSample, powers(binNum)'&
! , binNum, numDensity, kVals(binNum), fluxSample, powers(binNum)

  end function calcNumDensity

  !---------------------------------------------------------------------
  function calcCumFluxDensity(fluxSample, fluxes, powers&
  , offsets, multipliers) result(undetectableTotalFlux)
  !---------------------------------------------------------------------
  ! See theory described in calcAccumFluxCoeffs.
  !
  !---------------------------------------------------------------------
  ! Seems to test out ok, although returns values about 4-5 % higher than
  ! those estimated by running srcmap at various values of flux cutoff.
  ! IMS 2007-08-14.
  !---------------------------------------------------------------------

    implicit none

    real(single), intent(in) :: fluxSample,&
                                fluxes(:),&
                                powers(size(fluxes)),&
                                offsets(size(fluxes)),&
                                multipliers(size(fluxes))
    real(single)             :: undetectableTotalFlux

    integer(int32) :: binNum

    !-------------------------------------------------------------------
    ! Start

if (size(fluxes)<1) call error('bug',"")
    binNum = getBinNumber(fluxes, fluxSample) ! in ssclib/array_utils.

    if (binNum <= 0) then
      undetectableTotalFlux = 0.0
  return
    end if

    if (2.0 + powers(binNum) == 0.0)&
call error('notYetSupported','')

    undetectableTotalFlux = offsets(binNum)&
      + multipliers(binNum) * fluxSample**(2.0+powers(binNum))

  end function calcCumFluxDensity

  !---------------------------------------------------------------------
  subroutine calcCumFluxDensities(srcSpecTab, fluxSamples, cumFluxDensities)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)  :: srcSpecTab
    real(single), intent(in)  :: fluxSamples(:)
    real(single), intent(out) :: cumFluxDensities(size(fluxSamples))

    integer :: numFluxSamples, numModelKnees, i
    real(single), pointer :: modelFluxes(:), modelCumSkyDensity(:)
    real(single) :: hiPower
    real(single), allocatable ::&
      modelPowers(:), offsets(:), multipliers(:), kVals(:)

    !-------------------------------------------------------------------
    ! Start

    numFluxSamples = size(fluxSamples)

    ! Read the logN-logS distribution model from the template file:
    !
    call getLogNlogS(srcSpecTab, modelFluxes, modelCumSkyDensity&
      , hiPower)

    numModelKnees = size(modelFluxes)
    if (numModelKnees<1)&
call error('tooFewFluxAnchors',"")

    allocate(modelPowers(numModelKnees), offsets(numModelKnees))
    allocate(multipliers(numModelKnees), kVals(numModelKnees))
    call calcLogNLogSPowers(modelFluxes, modelCumSkyDensity, hiPower&
      , modelPowers)
    call calcKVals(modelFluxes, modelCumSkyDensity, modelPowers, kVals)
    call calcAccumFluxCoeffs(modelFluxes, kVals, modelPowers&
      , offsets, multipliers)

    do i = 1, numFluxSamples
      cumFluxDensities(i) = calcCumFluxDensity(fluxSamples(i), modelFluxes&
        , modelPowers, offsets, multipliers)
    end do

    deallocate(modelFluxes, modelCumSkyDensity)

  end subroutine calcCumFluxDensities

  !---------------------------------------------------------------------
  function calcReverseCumFluxDensity(fluxSample, fluxes, powers&
  , offsets, multipliers) result(totalFluxDensity)
  !---------------------------------------------------------------------
!*** not yet tested

    implicit none

    real(single), intent(in) :: fluxSample,&
                                fluxes(:),&
!                                cumSkyDensity(size(fluxes)),&
                                powers(size(fluxes)),&
                                offsets(size(fluxes)),&
                                multipliers(size(fluxes))
    real(single)             :: totalFluxDensity

    integer :: numFluxes, binNum

    !-------------------------------------------------------------------
    ! Start

    numFluxes = size(fluxes)
    if (numFluxes<1)&
call error('bug',"")

    binNum = getBinNumber(fluxes, fluxSample) ! in ssclib/array_utils.

    if (binNum <= 0) then
      totalFluxDensity = offsets(1)
    else if (binNum >= numFluxes) then
      totalFluxDensity = multipliers(numFluxes) * fluxSample**(2.0+powers(numFluxes))
    else
      totalFluxDensity = offsets(binNum+1)&
        + multipliers(binNum) * fluxSample**(2.0+powers(binNum))
    end if

  end function calcReverseCumFluxDensity

  !---------------------------------------------------------------------
  function calcReverseCumNumDensity(fluxSample, fluxes, kVals, powers)&
  result(totalNumDensity)
  !---------------------------------------------------------------------
!*** not yet tested

    implicit none

    real(single), intent(in) :: fluxSample,&
                                fluxes(:),&
                                kVals(size(fluxes)),&
                                powers(size(fluxes))
    real(single)             :: totalNumDensity

    integer :: numFluxes, binNum

    !-------------------------------------------------------------------
    ! Start

    numFluxes = size(fluxes)
    if (numFluxes<1)&
call error('bug',"")
    if (any(1+powers==0))&
call error('notYetSupported','')

    binNum = getBinNumber(fluxes, fluxSample) ! in ssclib/array_utils.

    if (binNum <= 0) then
      totalNumDensity = (-kVals(     1) / (1.0 + powers(     1)))&
        * fluxes( 1)**(1.0 + powers(     1))
    else
      totalNumDensity = (-kVals(binNum) / (1.0 + powers(binNum)))&
        * fluxSample**(1.0 + powers(binNum))
    end if

  end function calcReverseCumNumDensity

  !---------------------------------------------------------------------
  subroutine template
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start

  end subroutine template

end module power_law_utils
